package server

import (
	"context"
	"fmt"
	"sync"
	"time"

	logging "git.loafle.net/commons/logging-go"
)

type Server interface {
	ListenAndServe() error
	Shutdown(ctx context.Context) error
	ConnectionSize() int

	HandleConnection(servlet Servlet, servletCtx ServletCtx, conn *Conn)
	HandleRead(conn *Conn, stopChan <-chan struct{}, doneChan chan<- struct{}, readChan chan []byte)
	HandleWrite(conn *Conn, stopChan <-chan struct{}, doneChan chan<- struct{}, writeChan chan []byte)
}

type Servers struct {
	ServerHandler ServerHandler

	ServerCtx   ServerCtx
	Connections sync.Map
	StopChan    chan struct{}
	StopWg      sync.WaitGroup
}

func (s *Servers) Shutdown(ctx context.Context) error {
	if s.StopChan == nil {
		return fmt.Errorf("server must be started before stopping it")
	}
	close(s.StopChan)
	s.StopWg.Wait()

	s.ServerHandler.Destroy(s.ServerCtx)

	s.StopChan = nil

	return nil
}

func (s *Servers) ConnectionSize() int {
	var sz int
	s.Connections.Range(func(k, v interface{}) bool {
		sz++
		return true
	})
	return sz
}

func (s *Servers) ServerMessage(msg string) string {
	return fmt.Sprintf("Server[%s]: %s", s.ServerHandler.GetName(), msg)
}

func (s *Servers) HandleConnection(servlet Servlet, servletCtx ServletCtx, conn *Conn) {
	addr := conn.RemoteAddr()

	defer func() {
		if nil != conn {
			conn.Close()
		}
		servlet.OnDisconnect(servletCtx)
		logging.Logger().Infof(s.ServerMessage(fmt.Sprintf("Client[%s] has been disconnected", addr)))
		s.StopWg.Done()
	}()

	logging.Logger().Infof(s.ServerMessage(fmt.Sprintf("Client[%s] has been connected", addr)))

	s.Connections.Store(conn, true)
	defer s.Connections.Delete(conn)

	servlet.OnConnect(servletCtx, conn)
	conn.SetCloseHandler(func(code int, text string) error {
		logging.Logger().Debugf("close")
		return nil
	})

	stopChan := make(chan struct{})
	servletDoneChan := make(chan struct{})

	readChan := make(chan []byte)
	writeChan := make(chan []byte)

	readerDoneChan := make(chan struct{})
	writerDoneChan := make(chan struct{})

	go servlet.Handle(servletCtx, stopChan, servletDoneChan, readChan, writeChan)
	go s.HandleRead(conn, stopChan, readerDoneChan, readChan)
	go s.HandleWrite(conn, stopChan, writerDoneChan, writeChan)

	select {
	case <-readerDoneChan:
		close(stopChan)
		<-writerDoneChan
		<-servletDoneChan
	case <-writerDoneChan:
		close(stopChan)
		<-readerDoneChan
		<-servletDoneChan
	case <-servletDoneChan:
		close(stopChan)
		<-readerDoneChan
		<-writerDoneChan
	case <-s.StopChan:
		close(stopChan)
		<-readerDoneChan
		<-writerDoneChan
		<-servletDoneChan
	}
}

func (s *Servers) HandleRead(conn *Conn, stopChan <-chan struct{}, doneChan chan<- struct{}, readChan chan []byte) {
	defer func() {
		doneChan <- struct{}{}
	}()

	if 0 < s.ServerHandler.GetMaxMessageSize() {
		conn.SetReadLimit(s.ServerHandler.GetMaxMessageSize())
	}
	if 0 < s.ServerHandler.GetReadTimeout() {
		conn.SetReadDeadline(time.Now().Add(s.ServerHandler.GetReadTimeout()))
	}
	conn.SetPongHandler(func(string) error {
		conn.SetReadDeadline(time.Now().Add(s.ServerHandler.GetPongTimeout()))
		return nil
	})

	var (
		message []byte
		err     error
	)

	for {
		readMessageChan := make(chan struct{})

		go func() {
			_, message, err = conn.ReadMessage()
			close(readMessageChan)
		}()

		select {
		case <-s.StopChan:
			<-readMessageChan
			return
		case <-readMessageChan:
		}

		if nil != err {
			if IsUnexpectedCloseError(err, CloseGoingAway, CloseAbnormalClosure) {
				logging.Logger().Debugf(s.ServerMessage(fmt.Sprintf("Read error %v", err)))
			}
			return
		}

		readChan <- message
	}
}

func (s *Servers) HandleWrite(conn *Conn, stopChan <-chan struct{}, doneChan chan<- struct{}, writeChan chan []byte) {
	defer func() {
		doneChan <- struct{}{}
	}()

	ticker := time.NewTicker(s.ServerHandler.GetPingPeriod())
	defer func() {
		ticker.Stop()
	}()
	for {
		select {
		case message, ok := <-writeChan:
			if 0 < s.ServerHandler.GetWriteTimeout() {
				conn.SetWriteDeadline(time.Now().Add(s.ServerHandler.GetWriteTimeout()))
			}
			if !ok {
				conn.WriteMessage(CloseMessage, []byte{})
				return
			}

			w, err := conn.NextWriter(TextMessage)
			if err != nil {
				return
			}
			w.Write(message)

			if err := w.Close(); nil != err {
				return
			}
		case <-ticker.C:
			conn.SetWriteDeadline(time.Now().Add(s.ServerHandler.GetPingTimeout()))
			if err := conn.WriteMessage(PingMessage, nil); nil != err {
				return
			}
		case <-stopChan:
			return
		}
	}
}
